#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/input.h>
#include <rote/rote.h>
#include <signal.h>
#include <fcntl.h>

#include <st2205.h>

#include <gd.h>
#include <gdfontt.h>
#include <gdfonts.h>
#include <gdfontl.h>

// array that maps evdev keycode => linux key
typedef struct _keymap {
    char keyname[16];     // name of key
    int  ch;              // TERM=linux key code
    int  shift_ch;        // TERM=linux key code while SHIFT is pressed
    int  ctrl_ch;         // TERM=linux key code while CONTROL is pressed
    int  alt_ch;          // TERM=linux key code while ALT is pressed
} keymap;

#define MAX_EVKEYS 256
static keymap kmap[MAX_EVKEYS]={{"",0},{"",0},{"",0}};

static int kmap2int(char *str)
{
    int n;
    if (*str=='\'' || *str=='"') return str[1];
    if (*str=='0' && str[1]=='x')
        if (sscanf(str+2,"%x",&n)==1) 
            return n;
    if (*str>='0' && *str<='9')
       return atoi(str);
    if (*str) fprintf(stderr,"Warning: kmap2int(%s) - number conversion failed\n",str);
    return 0; 
}
void load_kmap(char *file)
{
    fprintf(stderr,"load_kmap(%s) - loading key defintions\n",file);
    FILE *f=fopen(file,"r");
    if (!f) { perror("fopen"); fprintf(stderr,"ERROR: could npt open file : %s\n",file); exit(1); }
    char line[256];
    int count=0;
    char keyname[16],code[8],base[8],shift[8],ctrl[8],alt[8];
    while (fgets(line,sizeof(line)-1,f))
    {
        if (sscanf(line,"%15s %7s %7s %7s %7s %7s",keyname,code,base,shift,ctrl,alt)==6)
        {
            if (keyname[0]=='#') continue;
            int k=kmap2int(code);
            if (k>=MAX_EVKEYS)
            { fprintf(stderr,"load_kmap(%s): Warning ignoring keycode out-of-range: %s",file,line); continue; }

            strcpy(kmap[k].keyname,keyname);
            kmap[k].ch=kmap2int(base);
            kmap[k].shift_ch=kmap2int(shift);
            kmap[k].ctrl_ch=kmap2int(ctrl);
            kmap[k].alt_ch=kmap2int(alt);
            count++;
        }
        else if (sscanf(line,"%1s",keyname)==1) // ignore blank lines
            fprintf(stderr,"load_kmap(%s): ERROR Syntax: %s",file,line);
    }
    fclose(f);
    fprintf(stderr,"load_kmap(%s) - %d key defintions loaded\n",file,count);
}

static int shift_active=0;
static int ctrl_active=0;
static int alt_active=0;

// return codes:
//  0 => no mapping
// -1 => special key (SHIFT,CTRL,ALT)
// >0 => Linux key code
int evkey2linuxkey(int code,int pressed)
{
    if (code>=MAX_EVKEYS) return 0;
    if (kmap[code].keyname[0]==0) return 0;

    if (strcmp(kmap[code].keyname,"SHIFT_LEFT")==0 || strcmp(kmap[code].keyname,"SHIFT_RIGHT")==0)
    { shift_active=pressed; return -1; }
    if (strcmp(kmap[code].keyname,"ALT_LEFT")==0 || strcmp(kmap[code].keyname,"ALT_RIGHT")==0)
    { alt_active=pressed; return -1; }
    if (strcmp(kmap[code].keyname,"CTRL")==0)
    { ctrl_active=pressed; return -1; }

    int linuxkey=0;
    if (pressed==0)
    {
        linuxkey=kmap[code].ch;
        if (alt_active) linuxkey=kmap[code].alt_ch;
        if (ctrl_active) linuxkey=kmap[code].ctrl_ch;
        if (shift_active) linuxkey=kmap[code].shift_ch;
    }
    return linuxkey;
}

int check_for_keyboard(char *kbdev)
{
    int fd=open(kbdev,O_RDONLY);
    if (fd>=0)
    {
        fcntl(fd,F_SETFL,O_NONBLOCK);
        char name[256]= "Unknown";
        if (ioctl(fd,EVIOCGNAME(sizeof(name)),name)>=0)
            fprintf(stderr,"Opened device %s => %s\n",kbdev,name);
    }
    return fd;
}

void draw_char(unsigned char *pixels,gdImagePtr im,int x,int y,int ch,int bg,int fg,gdFontPtr font,int sh,int sw)
{
    gdImageFilledRectangle(im,x,y,x+font->w-1,y+font->h-1,bg);
    gdImageChar(im,font,x,y,ch,fg);
    
    int m,n;
    unsigned int c;
    for (m=y;m<y+font->h;m++) {
        unsigned char *p=pixels+m*(sw+sw+sw)+x+x+x;
        for (n=x;n<x+font->w;n++) {
            c=gdImageGetPixel(im,n,m);
            *p++=gdImageRed(im,c);
            *p++=gdImageGreen(im,c);
            *p++=gdImageBlue(im,c);
        }
    }
}

int main(int argc, char **argv)
{
    int verbose=0;
    int fd=-1;
    gdFontPtr font=gdFontTiny;

    char *st2205dev=NULL;
    char *kbdev=NULL;
    char *keymapfile=NULL;

    int i,bad=0;
    for (i=1;i<argc;i++)
    {
        if (strcmp(argv[i],"-v")==0 || strcmp(argv[i],"--verbose")==0) { verbose=1; continue; }
        if (strcmp(argv[i],"--font=tiny")==0) { font=gdFontTiny; continue; }
        if (strcmp(argv[i],"--font=small")==0) { font=gdFontSmall; continue; }
        if (strcmp(argv[i],"--font=large")==0) { font=gdFontLarge; continue; }
        if (strncmp(argv[i],"--keymap=",9)==0) { keymapfile=argv[i]+9; continue; }
        if (strncmp(argv[i],"--picframe=",11)==0) { st2205dev=argv[i]+11; continue; }
        if (strncmp(argv[i],"--keyboard=",11)==0) { kbdev=argv[i]+11; continue; }
        fprintf(stderr,"unrecognized command line option '%s'\n",argv[i]);
    }
    if (st2205dev==NULL || kbdev==NULL || keymapfile==NULL) bad=1;
    if (bad)
    {
        fprintf(stderr,"usage: %s [--font=tiny|small|large] --keyboard=/dev/input/event0 "
                       "--picframe=/dev/sdb --keymap=us.kmap\n",argv[0]);
        exit(1);
    }

    // open picframe
    st2205_handle *h=st2205_open(st2205dev);
    if (!h) { fprintf(stderr,"ERROR: could not open st2205 device %s\n",st2205dev); exit(1); }
    int sw=h->width,sh=h->height;
    unsigned char *pixels=malloc(sw*sh*3);

    // create libgd image
    gdImagePtr im=gdImageCreate(sw,sh);
    int colors[8];
    int black=colors[0]=gdImageColorAllocate(im,0,0,0);
    int red=colors[1]=gdImageColorAllocate(im,255,0,0);
    int green=colors[2]=gdImageColorAllocate(im,0,255,0);
    int yellow=colors[3]=gdImageColorAllocate(im,255,255,0);
    int blue=colors[4]=gdImageColorAllocate(im,0,0,255);
    int magenta=colors[5]=gdImageColorAllocate(im,255,0,255);
    int cyan=colors[6]=gdImageColorAllocate(im,0,255,255);
    int white=colors[7]=gdImageColorAllocate(im,255,255,255);
    
    // create in memory terminal
    int tw=sw/font->w,th=sh/font->h;
    RoteTerm *rt=rote_vt_create(th,tw);
    pid_t pid=rote_vt_forkpty(rt, "/bin/bash --login");
    fprintf(stderr,"rote_vt_forkpty: PID=%d\n",pid);

    char *lastbuf=(char *)calloc(th*tw,1);
    char *lastattr=(char *)calloc(th*tw,1);
    int last_cy=0,last_cx=0;

    load_kmap(keymapfile);

    struct input_event ev[64];
    fd_set readset,*rs=NULL,exceptset,*es=NULL;
    for(;;)
    {
        if (fd<0) // no current keyboard => check if one was plugged in 
        {
            fd=check_for_keyboard(kbdev);
            if (fd>=0) { rs=&readset; FD_ZERO(rs); es=&exceptset; FD_ZERO(es); }
        }

        // wait 200ms for keystrokes, or 1sec if there is no keybaord
        struct timeval timeout={1,0};
        if (rs) { FD_SET(fd,rs); FD_SET(fd,es); timeout.tv_sec=0; timeout.tv_usec=200000; } 
        if (select(rs?fd+1:0,rs,NULL,es,&timeout)>0)
        {
            // something happened - keypress or keyboard disconnect
            int rb=read(fd,ev,sizeof(ev));
            if (rb<0 || FD_ISSET(fd,es)) 
            { 
                fprintf(stderr,"Keyboard unplugged\n");
                close(fd); fd=-1; rs=es=NULL; 
                continue; 
            }

            // map and forward all the keystrokes to the terminal
            if (rb<(int)sizeof(struct input_event)) { perror("evtest: short read"); exit (1); }
            int e;
            for (e=0;e<(int)(rb/sizeof(struct input_event));e++)
            {
                if (ev[e].type!=EV_KEY) continue;

                int code=ev[e].code;
                int pressed=ev[e].value;

                int linuxkey=evkey2linuxkey(code,pressed);

                if (verbose)
                {
                    char *kn="????";
                    if (code<MAX_EVKEYS) kn=kmap[code].keyname;
                    fprintf(stderr,"%ld.%06ld evdev keycode %d keyname=%s is %s\n",
                           ev[e].time.tv_sec,ev[e].time.tv_usec,code,kn,pressed?"down":"up");
                    switch (linuxkey)
                    {
                    case -1: fprintf(stderr,"    special key\n"); break;
                    case 0:  fprintf(stderr,"    no mapping\n"); break;
                    default: 
                        fprintf(stderr,"    maps => %s linux key code 0x%02x (%c)\n",
                               kmap[code].keyname,linuxkey,linuxkey>=32 && linuxkey<126?linuxkey:'?');
                        break;
                    }
                }
                if (linuxkey>0) rote_vt_keypress(rt,linuxkey); // send keystroke to terminal
            }
        }

        // redraw the terminal screen in memory
        rote_vt_draw(rt,NULL,1,1,NULL);

        // copy any changes to the picframe
        int i,j,ch,attr,x,y=0,count=0;
        char *b=lastbuf,*a=lastattr;
        int cy=rt->crow,cx=rt->ccol;
        for (i=0;i<th;i++)
        {
            x=0;
            for (j=0;j<tw;j++)
            {
                ch=rt->cells[i][j].ch;
                attr=rt->cells[i][j].attr;
                if (ch!=*b || attr!=*a)
                {
                    if (j==cx && i==cy)
                        draw_char(pixels,im,x,y,ch,ROTE_ATTR_FG(attr),ROTE_ATTR_BG(attr),font,sh,sw); // inverse video for cursor
                    else
                        draw_char(pixels,im,x,y,ch,ROTE_ATTR_BG(attr),ROTE_ATTR_FG(attr),font,sh,sw);
                    *b=ch; *a=attr;
                    count++;
                }
                b++; a++;
                x+=font->w;
            }
            y+=font->h;
        }
        if (cy!=last_cy || cx!=last_cx) // redraw cursor if it has moved
        {
            draw_char(pixels,im,cx*font->w,cy*font->h,rt->cells[cy][cx].ch,
                      ROTE_ATTR_FG(rt->cells[cy][cx].attr),ROTE_ATTR_BG(rt->cells[cy][cx].attr),
                      font,sh,sw); // inverse video for cursor
            draw_char(pixels,im,last_cx*font->w,last_cy*font->h,rt->cells[last_cy][last_cx].ch,
                      ROTE_ATTR_BG(rt->cells[cy][cx].attr),ROTE_ATTR_FG(rt->cells[cy][cx].attr),
                      font,sh,sw);
            last_cy=cy; last_cx=cx;
            count++;
        }
        if (count) // push any changes to picframe
        {
            if (h->oldpix) { free(h->oldpix); h->oldpix=NULL; } // avoid libst2205 incremental update bug
            st2205_send_data(h,pixels);
        }
    }
    return 0;
}
